/* 
  CPP_CONTEST=2015 
  CPP_PROBLEM=A 
  CPP_LANG=CPP+OPENMP+MPI 
  CPP_PROCESSES_PER_NODE=marte 1 
*/ 

#include <stdlib.h> 
#include <stdio.h> 
#include <omp.h> 
#include <mpi.h> 
#include <iostream>

using namespace std;

#define NUMHILOS1 6 //se establece aquí el número de threads, independiente del de tareas
#define NUMHILOS2 6 //y el número de hilos en la multiplicación de matrices

void escribirmatriz(double *a,int n)
{
  for(int i=0;i<n;i++)
  {
    for(int j=0;j<n;j++)
      cout << a[i*n+j] << " ";
    cout <<endl;
  }
  cout <<endl;
}

void multiplication(double *a,int n,int ld) 
{ 
  double *r=new double[n*n]; 
  int i;
  omp_set_num_threads(NUMHILOS2); //número de hilos de segundo nivel
  #pragma omp parallel for private(i) collapse(2)
  for(i=0;i<n;i++) 
  { 
    for(int j=0;j<n;j++) 
    { 
       double s=0.; 
       for(int k=0;k<n;k++) 
       { 
         s+=a[i*ld+k]*a[k*ld+j]; 
       } 
       r[i*n+j]=s; 
    } 
  } 
  
  for(int i=0;i<n;i++) 
  { 
    for(int j=0;j<n;j++) 
    { 
       a[i*ld+j]=r[i*n+j]; 
    } 
  } 

  delete [] r; 
} 

void primera(int N,int M,int L,int S,double *A,int *pasos,int tarea)
{
  for(int i=1;i<=S;i++) //por cada paso
  {
    int trabajar=0; //para indicar si puede hacer el trabajo, inicialmente no
    while(trabajar==0)
    {
      #pragma omp critical
        if(pasos[tarea+1]==(i-1)) //si el paso anterior del siguiente está hecho
          trabajar=1; //se puede hacer el paso actual
    }
    multiplication(A,M,N);
//    cout <<"despues de multiplicacion " << i << " de la tarea 0 "<<endl;
//    escribirmatriz(A,N);
    #pragma omp critical
        pasos[tarea]=i; //se apunta que se ha hecho ya ese paso
  }
}  

void ultima(int N,int M,int L,int S,double *A,int *pasos,int tarea)
{
  for(int i=1;i<=S;i++) //por cada paso
  {
    int trabajar=0; //para indicar si puede hacer el trabajo, inicialmente no
    while(trabajar==0)
    {
      #pragma omp critical
        if(pasos[tarea-1]==i) //si el anterior ya ha hecho este paso
          trabajar=1; //se puede hacer el paso actual
    }
    multiplication(A,M,N);
//    cout <<"despues de multiplicacion " << i << " de la tarea " << tarea<<endl;
//    escribirmatriz(&A[((M-L)*tarea*N+(M-L)*tarea)*(-1)],N);
    #pragma omp critical
        pasos[tarea]=i; //se apunta que se ha hecho ya ese paso
  }
}  

void intermedia(int N,int M,int L,int S,double *A,int *pasos,int tarea)
{
  for(int i=1;i<=S;i++) //por cada paso
  {
    int trabajar=0; //para indicar si puede hacer el trabajo, inicialmente no
    while(trabajar==0)
    {
      #pragma omp critical
        if(pasos[tarea-1]==i && pasos[tarea+1]==(i-1)) //si el anterior ya ha hecho este paso y el siguiente el anterior
          trabajar=1; //se puede hacer el paso actual
    }
    multiplication(A,M,N);
//    cout <<"despues de multiplicacion " << i << " de la tarea " << tarea<<endl;
//    escribirmatriz(&A[((M-L)*tarea*N+(M-L)*tarea)*(-1)],N);
    #pragma omp critical
        pasos[tarea]=i; //se apunta que se ha hecho ya ese paso
  }
}  

void update(int n,double *a,double d) 
{ 
  for(int i=0;i<n*n;i++) 
  { 
    a[i]/=d; 
  } 
} 

void  sec(int N,int M,int L,int S,double *A,int nodo,int np) 
{ 
  omp_set_nested(1); //se pone paralelismo anidado para que al hacer cada multiplicación también se use paralelismo
  omp_set_num_threads(NUMHILOS1); //búmero de hilos de primer nivel
  update(N,A,10.);
  int nummat=1+(N-M)/(M-L); //número de submatrices 
  int *pasos=new int[nummat]; //para almacenar el número de paso que se ha hecho de cada matriz
  for(int i=0;i<nummat;i++)
    pasos[i]=0; //inicialmente no se ha hecho ningún paso 
  #pragma omp parallel  //se ponen en marcha los hilos 
  {
//    cout <<"hilo "<<omp_get_thread_num()<<endl;
    #pragma omp single nowait//un único threads lanza las tareas
    {
      #pragma omp task
          primera(N,M,L,S,A,pasos,0); //la primera tarea hace las operaciones con la primera matriz
//      cout <<"lanzada tarea "<<0<<endl;
      #pragma omp task
          ultima(N,M,L,S,&A[(M-L)*(nummat-1)*N+(M-L)*(nummat-1)],pasos,nummat-1); //la última hace las operaciones con la última matriz
//     cout <<"lanzada tarea "<<nummat-1<<endl;
      for(int i=1;i<nummat-1;i++)
      {
        #pragma omp task firstprivate(i)
          intermedia(N,M,L,S,&A[(M-L)*i*N+(M-L)*i],pasos,i); //y las intermedias
//          cout <<"lanzada tarea "<<i<<endl;
      }
    }
  }
  delete[] pasos;
}
