/* 
  CPP_CONTEST=2015 
  CPP_PROBLEM=A 
  CPP_LANG=CPP+OPENMP+MPI 
  CPP_PROCESSES_PER_NODE=marte 1 mercurio 1
*/ 

#include <stdlib.h> 
#include <stdio.h> 
#include <omp.h> 
#include <mpi.h> 
#include <iostream>

using namespace std; 

void copiarAaAl(double *a,int M,int N,double *r)
{
  for(int i=0;i<M;i++)
    for(int j=0;j<M;j++)
      r[i*M+j]=a[i*N+j];
}

void copiarAlaA(double *r,int M,int N,double *a)
{
  for(int i=0;i<M;i++)
    for(int j=0;j<M;j++)
      a[i*N+j]=r[i*M+j];
}

void multiplication(double *a,int M,int N,int *tamano,int *inicio,int nodo,int np) 
{ 
  double *Al=new double[M*M]; //matriz local
  double *Rl; //matriz resultado local
  if(nodo==0)
    Rl=new double[M*M]; //en 0 completa, pues guarda el resultado final de Al*Al
  else
    Rl=new double[M*tamano[nodo]]; //en los dems slo las filas que calcula
  int i;
  MPI_Status *status;
  
  if(nodo==0)
    copiarAaAl(a,M,N,Al);
  MPI_Bcast(Al,M*M,MPI_DOUBLE,0,MPI_COMM_WORLD);
omp_set_num_threads(6);
#pragma omp parallel for private(i) collapse(2)
  for(i=inicio[nodo];i<inicio[nodo]+tamano[nodo];i++) //cada proceso calcula las filas que le corresponden
  { 
    for(int j=0;j<M;j++) 
    { 
       double s=0.; 
       for(int k=0;k<M;k++) 
       { 
         s+=Al[i*M+k]*Al[k*M+j]; 
       } 
       Rl[(i-inicio[nodo])*M+j]=s; 
    } 
  } 
  
//cout <<"proceso "<< nodo << " calculada matriz"<<endl;

  if(nodo==0)
  {
    for(int i=1;i<np;i++) //recibe punto a punto, podra ser con Gatherv
      MPI_Recv(&Rl[inicio[i]*M],tamano[i]*M,MPI_DOUBLE,i,10,MPI_COMM_WORLD,status);
  }
  else
    MPI_Send(Rl,tamano[nodo]*M,MPI_DOUBLE,0,10,MPI_COMM_WORLD);

  if(nodo==0) //deja el resultado en la zona correspondiente de la matriz completa
    copiarAlaA(Rl,M,N,a);

  delete[] Al;
  delete[] Rl; 
} 

void multiplications(int N,int M,int L,double *A,int * tamano,int *inicio,int nodo,int np) 
{ 
  for(int pos=0;pos<=N-M;pos+=M-L) 
  { 
    multiplication(&A[pos*N+pos],M,N,tamano,inicio,nodo,np); 
  //  cout <<"proceso "<< nodo << " posicion " << pos<<endl;
  } 
} 

void update(int n,double *a,double d) 
{ 
  for(int i=0;i<n*n;i++) 
  { 
    a[i]/=d; 
  } 
} 

void  sec(int N,int M,int L,int S,double *A,int nodo,int np) 
{ 
  int *datos=new int[4];//para guardar N, M, L y S
  if(nodo==0) //es el que tiene los datos y los pone en un array para mandarlos
  {
    datos[0]=N;
    datos[1]=M;
    datos[2]=L;
    datos[3]=S; //en vez de ponerlos en un array se podrían empaquetar para mandarlos
  }
  MPI_Bcast(datos,4,MPI_INT,0,MPI_COMM_WORLD);
  if(nodo!=0)
  {
    N=datos[0];
    M=datos[1];
    L=datos[2];
    S=datos[3]; 
  }

  int *tamano=new int[np],*inicio=new int[np]; //arrays de tamaños e inicios, que calculan todos
  int tambas=M/np; //el tamaño básico
  for(int i=0;i<np;i++)
  {
    tamano[i]=tambas;
    if(i<M%np) //el resto de filas asigna una más a los primeros procesos
      tamano[i]++;
  }
  inicio[0]=0;
  for(int i=1;i<np;i++)
  {
    inicio[i]=inicio[i-1]+tamano[i-1];
  }

  if(nodo==0) //es el que tiene la matriz, por lo que la actualiza sólo él
    update(N,A,10.); 

 //cout <<"proceso "<< nodo <<endl;
  for(int step=0;step<S;step++) 
  { 
     multiplications(N,M,L,A,tamano,inicio,nodo,np); //añadimos los parámetros de tamano, inicio, número de proceso y procesos totales 
   //  cout <<"proceso "<< nodo << "paso " << step<<endl;

  } 

  delete[] datos;
  delete[] tamano;
  delete[] inicio;
}
