/* 
  CPP_CONTEST=2015 
  CPP_PROBLEM=A 
  CPP_LANG=CPP+OPENMP+MPI 
  CPP_PROCESSES_PER_NODE=marte 3 mercurio 3
*/ 

#include <stdlib.h> 
#include <stdio.h> 
#include <omp.h> 
#include <mpi.h> 
#include <iostream>

using namespace std; 

void copiarAaAl(double *a,int M,int N,double *r)
{
  for(int i=0;i<M;i++)
    for(int j=0;j<M;j++)
      r[i*M+j]=a[i*N+j];
}

void copiarAlaA(double *r,int M,int N,double *a)
{
  for(int i=0;i<M;i++)
    for(int j=0;j<M;j++)
      a[i*N+j]=r[i*M+j];
}

void multiplicacionlocal(double *a,int n) 
{ 
  int i;
  double *r=new double[n*n];
  omp_set_num_threads(2);
#pragma omp parallel for private(i) collapse(2)
  for(i=0;i<n;i++) //cada proceso calcula las filas que le corresponden
  { 
    for(int j=0;j<n;j++) 
    { 
       double s=0.; 
       for(int k=0;k<n;k++) 
       { 
         s+=a[i*n+k]*a[k*n+j]; 
       } 
       r[i*n+j]=s; 
    } 
  } 
  copiarAaAl(r,n,n,a);

  delete[] r; 
} 

void update(int n,double *a,double d) 
{ 
  for(int i=0;i<n*n;i++) 
  { 
    a[i]/=d; 
  } 
} 

void  sec(int N,int M,int L,int S,double *A,int nodo,int np) 
{
    int *datos=new int[4];//para guardar N, M, L y S
    if(nodo==0) //es el que tiene los datos y los pone en un array para mandarlos
    {
      datos[0]=N;
      datos[1]=M;
      datos[2]=L;
      datos[3]=S; //en vez de ponerlos en un array se podran empaquetar para mandarlos
    }
    MPI_Bcast(datos,4,MPI_INT,0,MPI_COMM_WORLD);
    if(nodo!=0)
    {
      N=datos[0];
      M=datos[1];
      L=datos[2];
      S=datos[3]; 
    }
    int nummat=1+(N-M)/(M-L); //nmero de submatrices 
  if(nodo<nummat) //los dems procesos no trabajan
  {   
    double *Al=new double[M*M]; //Para almacenar la matriz local con la que trabaja cada proceso
    double *matext=new double[L*L]; //para copiar las matrices de los extremos al mandarlas
    MPI_Status *status;
  
    if(nodo==0) //es el que tiene la matriz, por lo que la actualiza slo l
      update(N,A,10.); 

   // cout <<"proceso "<< nodo << "tras actualizar " <<endl;
//envo a cada proceso de su matriz
    if(nodo==0)
    {
      for(int pos=M-L,i=1;pos<=N-M;pos+=M-L,i++) 
      { 
        copiarAaAl(&A[pos*N+pos],M,N,Al);  //antes de enviar, el cero tiene que copiar los datos a una zonz contigua de memoria
        MPI_Send(Al,M*M,MPI_DOUBLE,i,10,MPI_COMM_WORLD);
  //  cout <<"proceso "<< nodo << " posicion " << pos<<endl;
      }
      copiarAaAl(A,M,N,Al); //y copia la primera matriz a su zona local
    } 
    else
      MPI_Recv(Al,M*M,MPI_DOUBLE,0,10,MPI_COMM_WORLD,status);
    //Sustituimos los pasos por los procesos, y dentro de cada proceso se harn los pasos de forma distinta segn el proceso    
    /*for(int step=0;step<S;step++) 
    { 
      multiplications(N,M,L,A,tamano,inicio,nodo,np); //aadimos los parmetros de tamano, inicio, nmero de proceso y procesos totales 
   //  cout <<"proceso "<< nodo << "paso " << step<<endl;
    }*/ 

    if(nodo==0) 
    {
       multiplicacionlocal(Al,M); //en el primer paso computa y enva a P1
       copiarAaAl(&Al[(M-L)*M+M-L],L,M,matext); //antes de enviar tiene que copiar en una matriz con los datos contiguos
       MPI_Send(matext,L*L,MPI_DOUBLE,1,20,MPI_COMM_WORLD);
       for(int step=2;step<=S;step++) 
      {
         MPI_Recv(matext,L*L,MPI_DOUBLE,1,20,MPI_COMM_WORLD,status); //recibe de abajo
         copiarAlaA(matext,L,M,&Al[(M-L)*M+M-L]); //copia a la zona correspodiente de la matriz
         multiplicacionlocal(Al,M); //en el primer paso computa y enva a P1
         copiarAaAl(&Al[(M-L)*M+M-L],L,M,matext); //antes de enviar tiene que copiar en una matriz con los datos contiguos
         MPI_Send(matext,L*L,MPI_DOUBLE,1,20,MPI_COMM_WORLD);
      }
    }
    else if(nodo==nummat-1)
    {
       for(int step=1;step<S;step++) 
      {
         MPI_Recv(matext,L*L,MPI_DOUBLE,nummat-2,20,MPI_COMM_WORLD,status); //recibe de arriba
         copiarAlaA(matext,L,M,Al); //copia a la zona correspodiente de la matriz
         multiplicacionlocal(Al,M); 
         copiarAaAl(Al,L,M,matext); //antes de enviar tiene que copiar en una matriz con los datos contiguos
         MPI_Send(matext,L*L,MPI_DOUBLE,nummat-2,20,MPI_COMM_WORLD);
      }
      MPI_Recv(matext,L*L,MPI_DOUBLE,nummat-2,20,MPI_COMM_WORLD,status); //recibe de arriba
      copiarAlaA(matext,L,M,Al); //copia a la zona correspodiente de la matriz
      multiplicacionlocal(Al,M); //en el ltimo paso no enva la matriz de la esquina
    }
    else
    {
      MPI_Recv(matext,L*L,MPI_DOUBLE,nodo-1,20,MPI_COMM_WORLD,status); //recibe de arriba
      copiarAlaA(matext,L,M,Al); //copia a la zona correspodiente de la matriz
      multiplicacionlocal(Al,M); 
      copiarAaAl(Al,L,M,matext); //antes de enviar tiene que copiar en una matriz con los datos contiguos
      MPI_Send(matext,L*L,MPI_DOUBLE,nodo-1,20,MPI_COMM_WORLD); //enva arriba
      copiarAaAl(&Al[(M-L)*M+M-L],L,M,matext); //antes de enviar tiene que copiar en una matriz con los datos contiguos
      MPI_Send(matext,L*L,MPI_DOUBLE,nodo+1,20,MPI_COMM_WORLD);
      for(int step=2;step<S;step++) 
      {
         MPI_Recv(matext,L*L,MPI_DOUBLE,nodo-1,20,MPI_COMM_WORLD,status); //recibe de arriba
         copiarAlaA(matext,L,M,Al); //copia a la zona correspodiente de la matriz
         MPI_Recv(matext,L*L,MPI_DOUBLE,nodo+1,20,MPI_COMM_WORLD,status); //recibe de abajo
         copiarAlaA(matext,L,M,&Al[(M-L)*M+M-L]); //copia a la zona correspodiente de la matriz
         multiplicacionlocal(Al,M); 
         copiarAaAl(Al,L,M,matext); //antes de enviar tiene que copiar en una matriz con los datos contiguos
         MPI_Send(matext,L*L,MPI_DOUBLE,nodo-1,20,MPI_COMM_WORLD); //enva arriba
         copiarAaAl(&Al[(M-L)*M+M-L],L,M,matext); //antes de enviar tiene que copiar en una matriz con los datos contiguos
         MPI_Send(matext,L*L,MPI_DOUBLE,nodo+1,20,MPI_COMM_WORLD);
      }
      MPI_Recv(matext,L*L,MPI_DOUBLE,nodo-1,20,MPI_COMM_WORLD,status); //recibe de arriba
      copiarAlaA(matext,L,M,Al); //copia a la zona correspodiente de la matriz
      MPI_Recv(matext,L*L,MPI_DOUBLE,nodo+1,20,MPI_COMM_WORLD,status); //recibe de abajo
      copiarAlaA(matext,L,M,&Al[(M-L)*M+M-L]); //copia a la zona correspodiente de la matriz
      multiplicacionlocal(Al,M); 
      //no enva arriba en el ltimo paso
      copiarAaAl(&Al[(M-L)*M+M-L],L,M,matext); //antes de enviar tiene que copiar en una matriz con los datos contiguos
      MPI_Send(matext,L*L,MPI_DOUBLE,nodo+1,20,MPI_COMM_WORLD);
    }

    if(nodo==0)
    {
      copiarAlaA(Al,M,N,A);
      for(int i=1;i<nummat;i++)
      {
        MPI_Recv(Al,M*M,MPI_DOUBLE,i,30,MPI_COMM_WORLD,status);
        copiarAlaA(Al,M,N,&A[(M-L)*i*N+(M-L)*i]);
      }
    }
    else
      MPI_Send(Al,M*M,MPI_DOUBLE,0,30,MPI_COMM_WORLD);
    delete[] Al;
    delete[] matext;
  }
}
